import json
import time
import sys
import traceback

from gawseed.threatfeed.config import Config
import gawseed.threatfeed.loader as loader

class Datasource(Config):
    """Pulls enrichment data from another datasource based on a key
    ("match_key") in the row that was matched by the searcher.  It
    matches this key against a key from the datasource:
    "datasource_key.  When a match is found, data will be gathered
    backward and forward from that time based on the specified
    backward_time and forward_time values.  The set of rows matched
    will be returned as an array.
    """
    def __init__(self, conf, search_index, data, is_binary):
        super().__init__(conf)

        self.require(['datasource'])

        self._datasource = self.config('datasource',
                                       help="A dictionary describing a datasource to pull data from.")
        self._match_key = self.config('match_key', 'id_orig_h',
                                      help="The key to use when extracting search data from the matched row.")

        self._match_time_column = self.config('datasource_time_column', 'timestamp',
                                          help="The time column to use when extracting a match timestamp to search forward and backward from")
        
        self._datasource_key = self.config('datasource_key', 'id_orig_h',
                                           help="The key to use when searching through the datasource.")

        self._time_backward = self.config('time_backward', '2m', datatype='offset',
                                          help="The amount of time to get data from going backward from the match point")
        self._time_forward = self.config('time_forward', '2m', datatype='offset',
                                         help="The amount of time to get data from going forward from the match point")

        self._output_key = self.config('output_key', 'datasource',
                                       help="The output key to store the returned data in.")

    def gather(self, count, row, match):

        ds_config = self._datasource

        timeval = row[self._match_time_column]
        timestamp = self.parse_time(timeval)
        ds_config['begin_time'] = '@' + str(timestamp - self._time_backward)
        ds_config['end_time'] = '@' + str(timestamp + self._time_forward)

        conf = { loader.YAML_KEY: [{loader.DATASOURCE_KEY: ds_config}] }

        data_source = loader.create_instance(conf, loader.DATASOURCE_KEY)
        data_source.initialize()

        try:
            data_source.open()
        except Exception as e:
            print("end of file? " + str(e))
            return (self._output_key, []) # end of file

        # how we should do it eventually:
        conf = { loader.YAML_KEY: [{loader.SEARCHER_KEY: { 'module': 'ip',
                                                           'search_keys': [self._datasource_key]}}] }

        search_index = {row[self._match_key]: row}
        searcher = loader.create_instance(conf, loader.SEARCHER_KEY,
                                          [search_index, data_source, data_source.is_binary()])

        enrich_rows = []
        try:
            for finding in next(searcher):
                enrich_rows.append(finding[0])
        except Exception as e:
            print(traceback.format_exc())
            print("done searching? exception: " + str(e))

        return (self._output_key, enrich_rows)
                



"""Utilities for dynamically loading python modules for specific functions,
   parsing yaml config files and loading needed modules"""

import importlib
import re
import yaml

YAML_KEY = 'threat-search'
THREATSOURCE_KEY = 'threatsource'
DATASOURCE_KEY = 'datasource'
SEARCHER_KEY = 'searcher'
REPORTER_KEY = 'reporter'
ENRICHMENT_KEY = 'enrichments'
YAML_SECTIONS = [THREATSOURCE_KEY, DATASOURCE_KEY, SEARCHER_KEY, REPORTER_KEY, ENRICHMENT_KEY]

# load modules?
MODULE_XFORMS = {
    THREATSOURCE_KEY: {
        'kafka': 'gawseed.threatfeed.feeds.kafka.KafkaThreatFeed',
        'fsdb': 'gawseed.threatfeed.feeds.fsdb.FsdbThreatFeed',
    },
    DATASOURCE_KEY: {
        'fsdb': 'gawseed.threatfeed.datasources.fsdb.FsdbDataSource',
        'bro': 'gawseed.threatfeed.datasources.bro.BroDataSource',
        'kafka': 'gawseed.threatfeed.datasources.kafka.KafkaDataSource',
    },
    SEARCHER_KEY: {
        'ssh': 'gawseed.threatfeed.search.ssh.SSHSearch',
        'ip': 'gawseed.threatfeed.search.ip.IPSearch',
        'http': 'gawseed.threatfeed.search.http.HTTPSearch',
        'dns': 'gawseed.threatfeed.search.dns.DNSSearch',
        're': 'gawseed.threatfeed.search.re.RESearch',
    },
    REPORTER_KEY: {
        'dumper': 'gawseed.threatfeed.events.dumper.EventStreamDumper',
        'printer': 'gawseed.threatfeed.events.printer.EventStreamPrinter',
        'reporter': 'gawseed.threatfeed.events.reporter.EventStreamReporter',
    },
    ENRICHMENT_KEY: {
        'url': 'gawseed.threatfeed.enrichments.EnrichmentURL',
        'datasource': 'gawseed.threatfeed.enrichments.datasource.Datasource',
    },
}

named_entries = {}

def dump_config_options(debug=False):
    print("threat-search:")
    first_char = "-"
    for part in MODULE_XFORMS:
        print("  %s %s:" % (first_char, part))
        print("    # ---- %s modules and options" % (part))
        print("    # (pick one module)")
        for module in MODULE_XFORMS[part]:
            print("      module: %s" % (module))

            try:
                module = load_module_name(MODULE_XFORMS[part][module])
                # this forces a module to just dump out config settings to stdout
                if module.__doc__:
                    doc = module.__doc__
                    doc = re.sub("\n", "\n      # ", doc)
                    print("      #       " + doc)
                    
                if part == SEARCHER_KEY:
                    module({'dump_config': 1}, None, None, False)
                elif part == ENRICHMENT_KEY:
                    module({'dump_config': 1}, None, None, False)
                else:
                    module({'dump_config': 1})
            except Exception as e:
                print("      # couldn't get config for this")
                if debug:
                    print("      # " + str(e))
                    

            print("")
            first_char = " "
    exit()
    
def load_module_name(module_name):
    """Given a module name, split it in parts, load the module and find its child name
    within and return the final loaded object"""
    try:
        lastdotnum = module_name.rindex('.')
        modulename = module_name[:lastdotnum]   # just the module name
        class_name = module_name[lastdotnum+1:] # just the class init function name

        module = importlib.import_module(modulename)
    except Exception as e:
        print(e)
        raise ValueError("Error parsing/loading module/class name: " + module_name)

    if not hasattr(module, class_name):
        raise ValueError("Error finding class name '" + class_name + "' in module '" + module + "'")

    return getattr(module, class_name)

def load_yaml_config(config_stream):
    """Loads a yaml config from a specific stream and loads class definitions based on it."""

    conf = yaml.load(config_stream, Loader=yaml.FullLoader)
    load_class_config(conf[YAML_KEY])
    return conf

def load_class_config(threatconfs, sections=YAML_SECTIONS):
    """Loops through config and loads classes for each one into a 'class' key"""

    # load each section
    for threatconf in threatconfs:
        for section in sections:
            if section in threatconf: # some are optional
                parts = threatconf[section]
                if type(parts) != list:
                    parts = [parts]
                for part in parts:
                    module = part['module']
                    if module in MODULE_XFORMS[section]:
                        module = MODULE_XFORMS[section][module]
                    part['class'] = load_module_name(module)

    return threatconf

def maybe_save_entry(conf):
    """Remembers a named configuration template for use later"""
    global named_entries

    if 'name' not in conf:
        return

    entry_name = conf['name']
    named_entries[entry_name] = dict(conf)
    del named_entries[entry_name]['name']
    
def copy_entry(entry_name, conf):
    """Creates a new configuration dictionary based on a named template and override values."""
    if entry_name not in named_entries:
        return conf

    # clone the template source
    newconf = dict(named_entries[entry_name])

    # add/override other variables
    for item in conf:
        newconf[item] = conf[item]

    return newconf

def create_instance(conf, module_type, args=[], initialize=True):
    """Creates an instantiated instance of a threat-feed module"""

    # save this if requested
    maybe_save_entry(conf)
        
    # if they've requested a template
    if 'use' in conf:
        conf = copy_entry(conf['use'], conf)

    if 'class' not in conf[YAML_KEY][0][module_type]:
        load_class_config(conf[YAML_KEY], [module_type])
    
    obj = conf[YAML_KEY][0][module_type]['class']

    created = obj(conf[YAML_KEY][0][module_type], *args)

    if initialize:
        created.initialize()

    return created




from gawseed.threatfeed.config import Config

class DataSource(Config):
    def __init__(self, conf):
        super().__init__(conf)

        self._binary = self.config("binary", self.default_is_binary(),
                                   help="If the datasource is binary (bytes) data, set this to convert when necessary")

    def initialize(self):
        super().initialize()

    def default_is_binary(self):
        return False

    def is_binary(self):
        return self._binary

    def convert_row_to_utf8(self, row):
        if not self._binary:
            return row
            
        utf8_row = {}
        for item in row:
            if type(row[item]) == bytes:
                utf8_row[self.decode_item(item)] = row[item].decode()
            else:
                utf8_row[item] = row[item]
        return utf8_row

    def encode_item(self, item):
        """converts a string item to a utf-8 bytes version of it"""

        # don't encode if not needed
        if not self._binary:
            return item

        if type(item) == bytes:
            return item
        return bytes(item, 'utf-8')

    def decode_item(self, item):
        """Decodes a bytes object into a utf-8 string"""

        # don't decode if not needed
        if not self._binary:
            return item

        if type(item) == bytes:
            return item.decode()
        return item

    def encode_dict(self, old_dict):
        """Creates a new dict that contains both binary and string based indexes for all entries"""

        # don't encode if not needed
        if not self._binary:
            return old_dict

        new_dict = {}
        for key in old_dict:
            # stores both new binary key and the old
            new_dict[self.encode_item(key)] = old_dict[key]
            new_dict[key] = old_dict[key]
        return new_dict

    def encode_list(self, old_list):
        """Creates a new  list converted to binary if necessary"""

        # don't encode if not needed
        if not self._binary:
            return old_list

        new_list = []
        for key in old_list:
            # stores both new binary key and the old
            new_list.append(self.encode_item(key))
        return new_list

    def maybe_convert_token_to_binary(self, value):
        # don't encode if not needed
        if not self._binary:
            return value

        if type(value) != str:
            return value
        if self._binary_search:
            return bytes(value, 'utf-8')
        return value

    def maybe_convert_list_to_binary(self, values):
        # don't encode if not needed
        if not self._binary:
            return values

        new_list = []
        for value in values:
            new_list.append(self.maybe_convert_token_to_binary(value))
        return new_list
    
